#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#ifdef QCDOC
#include <qcdoc.h>
#include <qcdocos.h>
#include <ppc_lib.h>
#include <qalloc.h>
#endif
#include <math.h>
#include "wfm.h"
#include "wilson.h"
#include "timersub.h"

#define Ncomplex 2
#define Ncolour 3
#define Nspin2 2
#define Nspin4 4
#define Ncb 2
#define ND 4
#define Re 0
#define Im 1


#define   MYRAND_MAX        (2147483647)
static unsigned long long rand_seed;
void mysrand(unsigned int seed)
{
  rand_seed = seed;
}
 
int myrand(void)
{
  int r;
 
  rand_seed = rand_seed *  6364136223846793005LL + 1;
  return (int) ( ( rand_seed  >> 32 ) & MYRAND_MAX );
}
void cps_wilson_dslash(Float *chi_p_f, 
                       Float *u_p_f, 
                       Float *psi_p_f, 
                       int cb,
                       int dag,
                       Wilson *wilson_p);

void compare_result(Float *ans,Float *ref, int nword);
void compare_result(Float *ans,Float *ref, int nword)
{
  int i;
  for (i=0;i<nword;i++) {
    if ( fabs(ans[i]-ref[i])/fabs(ans[i]) > 5.0E-5 ) { 
      printf("Miscompare word %d %le != %le\n",i,ans[i],ref[i]);
     // exit(0);
    }
  }
  printf("result ok");
}

struct Spinor { 
  Float vals[ND][Ncolour][Ncomplex]; 
};
typedef struct Spinor Spin4;


typedef struct { Float vals[Nspin4][Ncolour][Ncolour][Ncomplex];} Gauge;

void RanFloat(Float *,int n);
void RanFloat(Float *f,int n)
{
  for ( int i=0;i<n;i++ ) {
    f[i] = (Float)myrand()/MYRAND_MAX;
  }
}

extern unsigned wfm_tbl1,wfm_tbl2,wfm_tbl3,wfm_tbl4;
extern "C" { 
  void bgl_mmu_optimise(void);
}

int main (int argc,char **argv )
{
#ifdef USE_COMMS_QMP
  QMP_thread_level_t thread_level_required=QMP_THREAD_SINGLE;
  QMP_thread_level_t thread_level_provided;
  QMP_init_msg_passing(&argc, &argv, thread_level_required, &thread_level_provided);
#endif

  int lx,ly,lz,lt;
  int lxyzt;
  int sx;

  FILE *fp;


  printf("Program started\n");

  //bgl_mmu_optimise();

  static int dimms[4];
#ifdef USE_COMMS_QMP
  if ( argc != 11 ) { 
    printf("Usage: %s lx ly lz lt sloppy bgl px py pz pt\n All must be even\n",argv[0]);
    exit(-1);
  }

  for(int i=0; i < 4; i++) { 
      dimms[i] = atoi(argv[7+i]);
  }
  QMP_status_t ltopP=QMP_declare_logical_topology(dimms, 4);
  if (ltopP != QMP_SUCCESS) { 
      fprintf(stderr, "Failed to declare logical topology mate \n");
      exit(-1);
  }

#else
  if ( argc != 7 ) { 
    printf("Usage: %s lx ly lz lt sloppy bgl\n All must be even\n",argv[0]);
    exit(-1);
  }
#endif
  lx = atoi(argv[1]);
  ly = atoi(argv[2]);
  lz = atoi(argv[3]);
  lt = atoi(argv[4]);

  int sloppy = atoi(argv[5]);
  int bgl    = atoi(argv[6]);

  if( (lx&0x1)||(ly&0x1)||(lz&0x1)||(lt&0x1) ) { 
    printf("Usage: %s lx ly lz lt\nAll must be even\n",argv[0]);
    printf("%d %d %d %d\n",lx,ly,lz,lt);
    fflush(stdout);
    exit(-1);
  }

  sx = lx/2;
  lxyzt = sx*ly*lz*lt;

  Spin4 *Chi  = (Spin4 *)ALLOC(sizeof(Spin4)*lxyzt*Ncb);
  Spin4 *PSI  = (Spin4 *)ALLOC(sizeof(Spin4)*lxyzt*Ncb);
  Spin4 *PSIck= (Spin4 *)ALLOC(sizeof(Spin4)*lxyzt*Ncb);
  Gauge *U    = (Gauge *)ALLOC(sizeof(Gauge)*lxyzt*Ncb);

  WilsonArg wil;
  Wilson    wilson_cpp;
  wfm mywfm;

  wil.local_latt[0]  = lx;
  wil.local_latt[1]  = ly;
  wil.local_latt[2]  = lz;
  wil.local_latt[3]  = lt;

  wil.local_comm[0]  = dimms[0]==1;
  wil.local_comm[1]  = dimms[1]==1;
  wil.local_comm[2]  = dimms[2]==1;
  wil.local_comm[3]  = dimms[3]==1;

  if ( bgl ) wil.BGL = true;
  else wil.BGL = false;

  if ( sloppy ) wil.SloppyPrecision = true;
  else wil.SloppyPrecision = false;

  wil.CoreCount = 1;

  wilson_cpp.ptr = (int *)malloc (4*sizeof(int));
  wilson_cpp.ptr[0] = lx;
  wilson_cpp.ptr[1] = ly;
  wilson_cpp.ptr[2] = lz;
  wilson_cpp.ptr[3] = lt;
  wilson_cpp.vol[0]  = (lx*ly*lz*lt)/2;
  wilson_cpp.vol[1]  = (lx*ly*lz*lt)/2;

  mywfm.init(&wil);
  double siteflops;

  for(int bench=0;bench <3;bench++){
  for(int i=0;i<1;i++ ) {

    printf("Iteration %d\n",i);
    RanFloat((Float *)Chi,lxyzt*24*Ncb);
    RanFloat((Float *)U,lxyzt*18*4*Ncb);
    RanFloat((Float *)PSI,lxyzt*24*Ncb);
    struct timeval t_start;
    struct timeval t_stop;

    int nmeas = 1000;
    gettimeofday(&t_start,NULL);


	switch (bench) {
	    case 0:
		for(int j=0;j<nmeas;j++){
		    mywfm.dslash((Float *)PSI,
        		       (Float *)U,
	    	               (Float *)Chi,0,0);
	            //exit(0);
                } 
		siteflops=1320;
		break;
	    case 1:
		for(int j=0;j<nmeas;j++)
		    mywfm.decom((Float *)PSI,
        		       (Float *)U,
	    	               0,0);
		siteflops=624;
		break;
	    case 2:
		for(int j=0;j<nmeas;j++)
		    mywfm.recon((Float *)Chi,
        		       (Float *)U,
	    	               0,0);
		siteflops=696;
		break;
	}
/*
      wfm_dslash_two( (Float *)PSI, (Float *)PSIck,
      		      (Float *)U,
		      (Float *)Chi, (Float *)Chi,
      		      0,0,0 );
*/

    gettimeofday(&t_stop,NULL);
    unsigned long t1,t2,t3,t4,t5;
/*
    t1=mfspr(sprg0);
    t2=mfspr(sprg1);
    t3=mfspr(sprg2);
    t4=mfspr(sprg3);
    t5=mfspr(usprg0);

    printf("Cycle counts:\n");
    printf("3xSu3: %d\n",t3-t1);
    printf("1xSu3: %d\n",t3-t2);
    printf("4rd Su3: %d\n",t4-t3);
    printf("rec    : %d\n",t5-t4);

*/
    timersub(&t_stop,&t_start,&t_start);
    double flops = lxyzt*siteflops *1.0 * nmeas;
    double time  = t_start.tv_sec + 1E-6*t_start.tv_usec;
    if ( i % 100 == 0  || 1)
      printf("%f Mflops\n",flops/time * 1.E-6);

  }
  }
    
  mywfm.dslash((Float *)PSI,
	       (Float *)U,
	       (Float *)Chi,0,0);

  cps_wilson_dslash((Float *)PSIck,
		    (Float *)U, 
		    (Float *)Chi,
		    0,
		    0,
		    &wilson_cpp);

  compare_result((Float *)PSI,(Float *)PSIck,lxyzt*24);

#ifdef USE_COMMS_QMP
  QMP_finalize_msg_passing();
#endif
  //  wfm_vec_end(&wil);
  exit(0);
}







