/****************************************************************************/
/* 10/16/97                                                                 */
/*                                                                          */
/* wilson.h                                                                 */
/*                                                                          */
/* C header file for the fermion wilson library wilson.lib                  */
/*                                                                          */
/****************************************************************************/

#ifndef INCLUDED_WILSON_H
#define INCLUDED_WILSON_H
#include <stdio.h>
#include <stdlib.h>
/*--------------------------------------------------------------------------*/
/* Definitions                                                              */
/*--------------------------------------------------------------------------*/
#define ND                 4      /* Space time dimension                   */
#define SPINOR_SIZE        24     /* Dirac spinor size                      */
#define HALF_SPINOR_SIZE   12     /* Half of the Dirac spinor size          */
#define PAD_HALF_SPINOR_SIZE 16     /* Half of the Dirac spinor size          */
#define BLOCK   HALF_SPINOR_SIZE  /* Size of spin projected 2 comp. spinor  */
#define COLUMN_SPINOR_SIZE  6     /* Size of one spinor component           */
#define GAUGE_SIZE         72     /* Gauge field size per site              */

/*--------------------------------------------------------------------------*/
/* External                                                                 */
/*--------------------------------------------------------------------------*/
extern int wfm_wire_map[];     
/* Set in wilson_int. For a given index 0-7 corresponding to
   X+, X-, Y+, Y-, Z+, Z-, T+, T-  it gives the corresponding wire. */

extern int wfm_max_scu_poll;
/* Set in wilson_init. The maximun number of times a wire is polled
   before wfm_scu_wait exits with an error. Implemented only in
   d_op_wilson_opt_nos_hdw_diag and d_op_dwf_nos_hdw_diag. */

extern int wfm_scu_diag[];
/* Contains diagnostic info about an scu failure to complete for
   wfm_max_scu_poll polls. Implemented only in 
   d_op_wilson_opt_nos_hdw_diag and d_op_dwf_nos_hdw_diag.
   The elements of the array are initialized to -1 (0xffffffff) in
   wilsonon_init (called by the lattice constructor of all wilson type 
   fermions) except for element 0 that is initialized to 0.
   The ellements of the array have the following meaning:
   0    :  Number of times an scu transfer has been initiated.
           Is set to 0 in wilson_init and is incremented every time 
           scu_comm_forward or scu_comm_backward are called.
   1    :  The value of the clock at exit. Set in wilson_scu_error.
   2    :  The value of the scu status register 813040. Set in wfm_scu_wait.
   3-10 :  The number of polls left for each of the 8 wires. Set in 
           wfm_scu_wait. It correspond to hardware wires 0-7.
   11-18:  The value of the 8 scu error status registers 813060-813067
           Set in wfm_scu_wait. It correspond to hardware wires 0-7. */

/*--------------------------------------------------------------------------*/
/* Structures                                                               */
/*--------------------------------------------------------------------------*/
/* This field makes one word = numblk[10bits]-blklen[10bits]-stride[12bits] */
struct comm_params
{
  unsigned int stride : 12;
  unsigned int blklen : 10;
  unsigned int numblk : 10;
};


/*--------------------------------------------------------------------------*/
/* Type definitions                                                         */
/*--------------------------------------------------------------------------*/
typedef double Float;
typedef struct{
   unsigned long *ptr;
   unsigned long *shift_table[2][2];   /*pointer to an array with addressing offsets */
   unsigned long *face_table[2][2][ND];/*pointer to an array with addressing offsets */
 
   /* temp spinor needed by mdagm                 */
   Float *spinor_tmp; 
   /* point. array to 4 interleaved fwd proj half spinors  */
   Float *spinf;
   /* point. array to 4 interleaved bwd proj half spinors  */   
   Float *spinb;
   Float *send_f[ND];
   Float *recv_f[ND];
   Float *send_b[ND];
   Float *recv_b[ND];
   int vol[2];
   int local_halfvol;
   int local_latt[ND];
   int local_comm[ND];
   int nbound[ND];
   int allbound;
  //   SCUDirArgMulti *comm_f;
  //   SCUDirArgMulti *comm_b;
} WilsonQCDOC;


/* The Wilson structure typedef */
typedef struct{
   int   *ptr;               /* pointer to an array with addressing offsets */
   int   yztmax;             /* # of points of the y-z-t lattice per node   */
   int   offset;             /* communication addressing related            */
   int   comm_offset[ND];    /* communication addressing related            */
   int   comm_stride[ND];    /* communication addressing related            */
   int   comm_blklen[ND];    /* communication addressing related            */
   int   comm_numblk[ND];    /* communication addressing related            */
   struct  comm_params comm[ND];  
   int   vol[2];
   int   padded_subgrid_vol[ND];
   Float *spinor_tmp;        /* temp spinor needed by mdagm                 */
   Float *af[ND];     /* point. array to 4 fwd spin projected half spinors  */
   Float *ab[ND];     /* point. array to 4 bwd spin projected half spinors  */
} WilsonQCDSP;

typedef WilsonQCDSP Wilson;



/*--------------------------------------------------------------------------*/
/* Function declarations                                                    */
/*--------------------------------------------------------------------------*/


/*==========================================================================*/
/* This routine performs all initializations needed before wilson library   */
/* funcs are called. It sets the addressing related arrays and reserves     */
/* memory for the needed temporary buffers. It only needs to be called only */
/* once at the begining of the program (or after a wilson_end call)         */
/* before any number of calls to wilson functions are made.                 */
/*==========================================================================*/
void wilson_init(Wilson *wilson_p);     /* pointer to Wilson struct. */


/*==========================================================================*/
/* This routine frees any memory reserved by wilson_init                    */
/*==========================================================================*/
void wilson_end(Wilson *wilson_p);     /* pointer to Wilson struct. */



/*==========================================================================*/
/* The Wilson fermion matrix * vector routines                              */
/* They are external "C"                                                    */
/*==========================================================================*/

extern "C"{

void wilson_mdagm(Float  *chi,        /* chi = MdagM(u) psi          */
		  Float  *u,          /* Gauge field                 */
		  Float  *psi,        /* chi = MdagM(u) psi          */
		  Float  *mp_sq_p,    /* pointer to Sum |M psi|^2    */
		  Float  Kappa,       /* Wilson's kappa parameter    */
		  Wilson *wilson_p);  /* pointer to a Wilson struct. */

void wilson_dslash(Float *chi, 
		   Float *u, 
		   Float *psi, 
		   int cb,
		   int dag,
		   Wilson *wilson_p);

void wilson_m(Float *chi, 
	      Float *u, 
	      Float *psi, 
	      Float kappa,
	      Wilson *wilson_p);

void wilson_mdag(Float *chi, 
		 Float *u, 
		 Float *psi, 
		 Float kappa,
		 Wilson *wilson_p);

/*QCDOC comms routines*/
void wfm_scatter_face(Wilson *wilson_p, int pm, int cb);
void wfm_comm_init(Wilson *wp);
void wfm_comm_forward_start(Wilson *wp);
void wfm_comm_backward_start(Wilson *wp);
void wfm_comm_forward_complete(Wilson *wp);
void wfm_comm_backward_complete(Wilson *wp);
                                                                                                   
}

#endif
